//
// Network Calculator
// Copyright (C) 2000 Mark Rigby-Jones <mrj@compsoc.net>
// http://www.compsoc.net/~mrj/palm/netcalc/
// 
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

#include <Pilot.h>
#include "netcalc.h"
#include "netcalcRsc.h"

Boolean error;
DmOpenRef database;
Char fieldoutput[SizeOutput], fieldaddrs[SizeAddrs];
Char fieldipaddr[SizeIPAddr], fieldipaddri[SizeIPAddr];
Char fieldnetmask[SizeIPAddr], fieldnetmaski[SizeIPAddr];
Char fieldnetwork[SizeIPAddr], fieldnetworki[SizeIPAddr];
Char fieldbroadcast[SizeIPAddr], fieldbroadcasti[SizeIPAddr];

static void CalculatePutNum(CharPtr buf, ULong ip) {
    if (ip > TenExFive) {
	StrIToA(buf, ip / TenExFive);
	buf += StrLen(buf);
	ip %= TenExFive;
    }
    StrIToA(buf, ip);
}

static void CalculatePutIP(CharPtr buf, ULong ip) {
    StrIToA(buf, ip / TwoExTwentyFour);
    ip = ip % TwoExTwentyFour;
    buf = buf + StrLen(buf);
    buf[0] = '.';
    buf++;
    StrIToA(buf, ip / TwoExSixteen);
    ip = ip % TwoExSixteen;
    buf = buf + StrLen(buf);
    buf[0] = '.';
    buf++;
    StrIToA(buf, ip / TwoExEight);
    buf = buf + StrLen(buf);
    buf[0] = '.';
    buf++;
    StrIToA(buf, ip % TwoExEight);
}

static ULong CalculateGetNum(CharPtr ptr) {
    ULong onum, num = 0;
    
    if (ptr == NULL) {
	error = true;
    } else if (ptr[0] == 0) {
	error = true;
    } else {
	while ((error == false) && (ptr[0] != 0)) {
	    if ((ptr[0] < '0') || (ptr[0] > '9')) {
		error = true;
	    } else {
		onum = num;
		num = (num * 10) + (ptr[0] - '0');
		if (num < onum) {
		    error = true;
		}
	    }
	    ptr++;
	}
    }

    return error ? 0 : num;
}

static ULong CalculateGetIP(CharPtr ptr) {
    CharPtr ptrdot;
    ULong ipaddr, tmp;

    ptrdot = StrChr(ptr, '.');
    if (ptrdot == NULL) {
	ipaddr = CalculateGetNum(ptr);
    } else {
        ptrdot[0] = 0;
        ipaddr = CalculateGetNum(ptr);
        ptrdot[0] = '.';
        ptr = ptrdot + 1;
        ptrdot = StrChr(ptr, '.');
        if ((ptrdot == NULL) || (ipaddr > 255)) {
            error = true;
        } else {
	    ptrdot[0] = 0;
	    tmp = CalculateGetNum(ptr);
	    ptrdot[0] = '.';
	    ptr = ptrdot + 1;
	    ptrdot = StrChr(ptr, '.');
	    if ((ptrdot == NULL) || (tmp > 255)) {
	        error = true;
	    } else {
	        ipaddr = (ipaddr * TwoExEight) + tmp;
	        ptrdot[0] = 0;
	        tmp = CalculateGetNum(ptr);
	        ptrdot[0] = '.';
	        if (tmp > 255) {
		    error = true;
	        } else {
		    ipaddr = (ipaddr * TwoExEight) + tmp;
		    tmp = CalculateGetNum(ptrdot + 1);
		    if (tmp > 255) {
		        error = true;
		    } else {
		        ipaddr = (ipaddr * TwoExEight) + tmp;
		    }
		}
	    }
	}
    }

    return error ? 0 : ipaddr;
}

static void Calculate(Boolean showerrs) {
    FormPtr frm;
    FieldPtr fldinput, fldoutput, fldaddrs;
    FieldPtr fldipaddr, fldnetmask, fldnetwork, fldbroadcast;
    FieldPtr fldipaddri, fldnetmaski, fldnetworki, fldbroadcasti;
    CharPtr ptr;
    Char sep = 0, input[SizeInput];
    ULong ipaddr, netmask = 32, network, broadcast, addrs, tmp;
    Short i;

    frm = FrmGetActiveForm();
    fldinput = FrmGetObjectPtr(frm,
	    FrmGetObjectIndex(frm, FieldInput));
    fldoutput = FrmGetObjectPtr(frm,
	    FrmGetObjectIndex(frm, FieldOutput));
    fldaddrs = FrmGetObjectPtr(frm,
	    FrmGetObjectIndex(frm, FieldAddrs));
    fldipaddr = FrmGetObjectPtr(frm,
	    FrmGetObjectIndex(frm, FieldIPAddr));
    fldnetmask = FrmGetObjectPtr(frm,
	    FrmGetObjectIndex(frm, FieldNetmask));
    fldnetwork = FrmGetObjectPtr(frm,
	    FrmGetObjectIndex(frm, FieldNetwork));
    fldbroadcast = FrmGetObjectPtr(frm,
	    FrmGetObjectIndex(frm, FieldBroadcast));
    fldipaddri = FrmGetObjectPtr(frm,
	    FrmGetObjectIndex(frm, FieldIPAddrI));
    fldnetmaski = FrmGetObjectPtr(frm,
	    FrmGetObjectIndex(frm, FieldNetmaskI));
    fldnetworki = FrmGetObjectPtr(frm,
	    FrmGetObjectIndex(frm, FieldNetworkI));
    fldbroadcasti = FrmGetObjectPtr(frm,
	    FrmGetObjectIndex(frm, FieldBroadcastI));
    StrNCopy(input, FldGetTextPtr(fldinput), SizeInput);

    error = false;
    
    if (input == NULL) {
	error = true;
    } else {
	ptr = StrChr(input, '/');
	if (ptr == NULL) {
	    ptr = StrChr(input, '-');
	}
	if (ptr == NULL) {
	    ptr = StrChr(input, '+');
	}
	if (ptr != NULL) {
	    sep = ptr[0];
	    ptr[0] = 0;
	}

	if (ptr == input) {
	    ipaddr = 0;
	} else {
	    ipaddr = CalculateGetIP(input);
	}

	if (ptr == NULL) {
	    if (ipaddr < (ULong)ClassA) {
		netmask = 8;
	    } else if (ipaddr < (ULong)ClassB) {
		netmask = 16;
	    } else {
		netmask = 24;
	    }
	} else {
	    ptr[0] = sep;
	    ptr++;
	    if (sep == '/') {
	    	netmask = CalculateGetIP(ptr);
	    	if (netmask > 32) {
		    tmp = -netmask;
		    for (netmask = 32; !(tmp % 2); netmask--) {
			tmp /= 2;
		    }
		    if (tmp != 1) {
		    	error = true;
		    }
	    	}
	    } else if (sep == '-') {
		broadcast = CalculateGetIP(ptr);
		if (broadcast < ipaddr) {
		    error = true;
		} else {
		    netmask = 32;
		    for (tmp = 1; (tmp > 0) && (broadcast >
			    (((ipaddr - (ipaddr % tmp)) + tmp) - 1));
			    tmp *= 2) {
			netmask--;
		    }
		}
	    } else if (sep == '+') {
		network = CalculateGetNum(ptr) + 2;
		if (network == 0) {
		    netmask = 0;
		} else if (network <= 2) {
		    error = true;
		} else {    
		    netmask = 31;
		    for (tmp = 2; (tmp > 0) && (tmp < network); tmp *= 2) {
		        netmask--;
		    }
		}
	    } else {
		netmask = 0;
	    }
	}
    }

    if (error == false) {
	CalculatePutIP(fieldoutput, ipaddr);
	ptr = fieldoutput + StrLen(fieldoutput);
	ptr[0] = '/';
	StrIToA(ptr + 1, netmask);
	if (netmask == 0) {
	    addrs = -2;
	    network = 0;
	    broadcast = -1;
	} else {
	    tmp = 1;
	    for (i = netmask; i < 32; i++) {
		tmp *= 2;
	    }
	    if (netmask > 30) {
		addrs = 0;
	    } else {
		addrs = tmp - 2;
	    }
	    netmask = -tmp;
	    network = ipaddr - (ipaddr % tmp);
	    broadcast = (network + tmp) - 1;
	}
	CalculatePutNum(fieldaddrs, addrs);
	ptr = fieldaddrs + StrLen(fieldaddrs);
	StrCat(ptr, " addrs");
	CalculatePutIP(fieldipaddr, ipaddr);
	CalculatePutIP(fieldnetmask, netmask);
	CalculatePutIP(fieldnetwork, network);
	CalculatePutIP(fieldbroadcast, broadcast);
	CalculatePutNum(fieldipaddri, ipaddr);
	CalculatePutNum(fieldnetmaski, netmask);
	CalculatePutNum(fieldnetworki, network);
	CalculatePutNum(fieldbroadcasti, broadcast);
    } else {
	fieldoutput[0] = 0;
	fieldaddrs[0] = 0;
	fieldipaddr[0] = 0;
	fieldnetmask[0] = 0;
	fieldnetwork[0] = 0;
	fieldbroadcast[0] = 0;
	fieldipaddri[0] = 0;
	fieldnetmaski[0] = 0;
	fieldnetworki[0] = 0;
	fieldbroadcasti[0] = 0;
	if (showerrs) {
	    FrmAlert(AlertFormatError);
	}
    }

    FldSetTextPtr(fldaddrs, fieldaddrs);
    FldRecalculateField(fldaddrs, true);
    FldSetTextPtr(fldoutput, fieldoutput);
    FldRecalculateField(fldoutput, true);
    FldSetTextPtr(fldipaddr, fieldipaddr);
    FldRecalculateField(fldipaddr, true);
    FldSetTextPtr(fldnetmask, fieldnetmask);
    FldRecalculateField(fldnetmask, true);
    FldSetTextPtr(fldnetwork, fieldnetwork);
    FldRecalculateField(fldnetwork, true);
    FldSetTextPtr(fldbroadcast, fieldbroadcast);
    FldRecalculateField(fldbroadcast, true);
    FldSetTextPtr(fldipaddri, fieldipaddri);
    FldRecalculateField(fldipaddri, true);
    FldSetTextPtr(fldnetmaski, fieldnetmaski);
    FldRecalculateField(fldnetmaski, true);
    FldSetTextPtr(fldnetworki, fieldnetworki);
    FldRecalculateField(fldnetworki, true);
    FldSetTextPtr(fldbroadcasti, fieldbroadcasti);
    FldRecalculateField(fldbroadcasti, true);
    FrmDrawForm(frm);
}

static Boolean MainFormHandleEvent(EventPtr e) {
    Boolean handled = false;
    FormPtr frm;
    FieldPtr fld, act;
    Word index;
    
    switch (e->eType) {
	case frmOpenEvent:
	    frm = FrmGetActiveForm();
	    index = FrmGetObjectIndex(frm, FieldInput);

    	    FldSetTextHandle(FrmGetObjectPtr(frm, index),
    		    (Handle)DmGetRecord(database, 0));
    	    FrmDrawForm(frm);
    	    FrmSetFocus(frm, index);

    	    Calculate(false);

    	    handled = true;
    	    break;

	case menuEvent:
	    MenuEraseStatus(NULL);
	    frm = FrmGetActiveForm();
	    index = FrmGetObjectIndex(frm, FieldInput);
	    fld = FrmGetObjectPtr(frm, index);
	    act = FrmGetObjectPtr(frm, FrmGetFocus(frm));

    	    switch(e->data.menu.itemID) {
    		case MenuitemMainRecordCalculate:
    		    Calculate(true);
    		    break;
    		case MenuitemMainRecordNew:
    		    FldDelete(fld, 0, FldGetTextLength(fld));
    		    FrmSetFocus(frm, index);
    		    Calculate(false);
    		    break;
    		case MenuitemMainEditUndo:
    		    if (fld != act) {
    			SndPlaySystemSound(sndError);
    		    } else {
    			FldUndo(fld);
    		    }
    		    break;
    		case MenuitemMainEditCopy:
    		    FldCopy(act);
    		    break;
    		case MenuitemMainEditCut:
    		    if (fld != act) {
    			SndPlaySystemSound(sndError);
    		    } else {
    			FldCut(fld);
    		    }
    		    break;
    		case MenuitemMainEditPaste:
    		    if (fld != act) {
    			SndPlaySystemSound(sndError);
    		    } else {
    			FldPaste(fld);
    		    }
		    break;
		case MenuitemMainEditSelect:
		    FldSetSelection(act, 0, FldGetTextLength(act));
		    break;
		case MenuitemMainEditKeyboard:
		    if (fld != act) {
			SndPlaySystemSound(sndError);
		    } else {
			SysKeyboardDialog(kbdNumbersAndPunc);
		    }
		    break;
		case MenuitemMainEditGraffiti:
		    SysGraffitiReferenceDialog(referenceAlpha);
		    break;
		case MenuitemMainOptionsHelp:
		    FrmHelp(StringHelp);
		    break;
		case MenuitemMainOptionsAbout:
		    FrmAlert(AlertAbout);
		    break;
	    }

	    handled = true;
	    break;

	case ctlSelectEvent:
	    switch(e->data.ctlSelect.controlID) {
		case ButtonCalculate:
		    Calculate(true);
		    break;
	    }
	    break;

	default:
	    break;
    }

    return handled;
}

static Boolean ApplicationHandleEvent(EventPtr e) {
    FormPtr frm;
    Word    formId;
    Boolean handled = false;

    if (e->eType == frmLoadEvent) {
	formId = e->data.frmLoad.formID;
	frm = FrmInitForm(formId);
	FrmSetActiveForm(frm);

	switch(formId) {
	case FormMain:
	    FrmSetEventHandler(frm, MainFormHandleEvent);
	    break;
	}
	handled = true;
    }

    return handled;
}

static Word StartApplication(void) {
    UInt index = 0;
    VoidHand handle;
    VoidPtr pointer;
    Char nullptr = 0;
    
    database = DmOpenDatabaseByTypeCreator(DataDBID, CreatorID,
	    dmModeReadWrite);
    if (!database) {
	if (DmCreateDatabase(0, DataDBName, CreatorID, DataDBID, false)) {
	    return 1;
	}
	database = DmOpenDatabaseByTypeCreator(DataDBID, CreatorID,
		dmModeReadWrite);
	handle = DmNewRecord(database, &index, 1);
	pointer = MemHandleLock(handle);
	DmWrite(pointer, 0, &nullptr, 1);
	MemPtrUnlock(pointer);
	DmReleaseRecord(database, index, true);
    }
	
    FrmGotoForm(FormMain);
    return 0;
}

static void StopApplication(void) {
    FormPtr frm;
    
    frm = FrmGetFormPtr(FormMain);
    FldSetTextHandle(FrmGetObjectPtr(frm,
		FrmGetObjectIndex(frm, FieldInput)), NULL);
    DmReleaseRecord(database, 0, true);
    DmCloseDatabase(database);
    FrmSaveAllForms();
    FrmCloseAllForms();
}

static void EventLoop(void) {
    Word err;
    EventType e;

    do {
	EvtGetEvent(&e, evtWaitForever);
	if (! SysHandleEvent (&e)) {
	    if (! MenuHandleEvent (NULL, &e, &err)) {
		if (! ApplicationHandleEvent (&e)) {
		    FrmDispatchEvent (&e);
		}
	    }
	}
    } while (e.eType != appStopEvent);
}

DWord PilotMain(Word cmd, Ptr cmdPBP, Word launchFlags) {
    Word err;
    DWord romversion;

    if (cmd == sysAppLaunchCmdNormalLaunch) {
	FtrGet(sysFtrCreator, sysFtrNumROMVersion, &romversion);
	if (romversion < ROMVersion20) {
	    return 1;
	}

	err = StartApplication();
	if (err) {
	    return err;
	}

	EventLoop();
	StopApplication();
    } else {
	return sysErrParamErr;
    }

    return 0;
}
